import tkinter
import time
from srcgen import statechart

import sys, os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))
from yakindu.rx import Observer


TIME_SCALE = 1

def scaled_now():
    return time.perf_counter_ns() * TIME_SCALE

class QueueEntry:
    __slots__ = ('timestamp', 'callback', 'canceled', 'debug') # For MAXIMUM performance :)

    def __init__(self, timestamp, callback, debug):
        self.timestamp = timestamp
        self.callback = callback
        self.debug = debug
        self.canceled = False

# Simulation primitive.
# Uses virtualized (simulated) time, instead of looking at the wall clock.
class Controller:
    def __init__(self):
        self.event_queue = []
        self.simulated_time = 0

    def start(self, timestamp):
        self.simulated_time = timestamp
        self.start_time = timestamp

    def add_input(self, timestamp, sc, event, value=None):
        raise_method = getattr(sc, 'raise_' + event)
        if value is not None:
            callback = lambda: raise_method(value)
        else:
            callback = raise_method
        print("input event: ", event)
        self.add_input_lowlevel(timestamp, callback, event)

    def add_input_lowlevel(self, timestamp, callback, debug):
        e = QueueEntry(timestamp, callback, debug)
        self.event_queue.append(e)
        self.event_queue.sort(key = lambda entry: entry.timestamp, reverse=True)
        return e

    def run_until(self, until):
        while self.have_event() and self.get_earliest() <= until:
            e = self.event_queue.pop();
            if not e.canceled:
                # print("handling", e.debug)
                self.simulated_time = e.timestamp
                e.callback()

    def have_event(self):
        return len(self.event_queue) > 0

    def get_earliest(self):
        return self.event_queue[-1].timestamp

# Our own timer service, used by the statechart.
# Much better than YAKINDU's pathetic timer service.
class TimerService:
    def __init__(self, controller):
        self.controller = controller;
        self.timers = {}

    # Duration: milliseconds
    def set_timer(self, sc, event_id, duration, periodic):
        def callback():
            sc.time_elapsed(event_id)

        self.unset_timer(callback, event_id)

        controller_duration = duration * 1000000 # ms to ns

        # print("set timer"+str(event_id), "duration", duration)
        e = self.controller.add_input_lowlevel(
            self.controller.simulated_time + controller_duration, # timestamp relative to simulated time
            callback,
            "timer"+str(event_id))

        self.timers[event_id] = e

    def unset_timer(self, callback, event_id):
        try:
            e = self.timers[event_id]
            e.canceled = True
        except KeyError:
            pass

# Runs the Controller as close as possible to the wall-clock.
# Depending on how fast your computer is, simulated time will always run a tiny bit behind wall-clock time, but this error will NOT grow over time.
class RealTimeSimulation:
    def __init__(self, Controller, tk, update_callback, scale=1.0):
        self.controller = controller
        self.tk = tk
        self.update_callback = update_callback
        self.scale = scale

        self.scheduled_id = None
        self.purposefully_behind = 0

    def add_input(self, sc, event, value=None):
        now = scaled_now() + self.purposefully_behind
        self.controller.add_input(now, sc, event, value)
        self.interrupt()

    def interrupt(self):
        if self.scheduled_id is not None:
            self.tk.after_cancel(self.scheduled_id)

        if self.controller.have_event():
            now = scaled_now()
            earliest = self.controller.get_earliest()
            sleep_time = (earliest - now) // TIME_SCALE

            if sleep_time < 0:
                self.purposefully_behind = sleep_time
            else:
                self.purposefully_behind = 0

            def callback():
                # print("run_until...")
                self.controller.run_until(now + self.purposefully_behind)
                self.update_callback()
                self.interrupt()

            # print("sleeping for", int(sleep_time / 1000000), "ms")

            self.scheduled_id = self.tk.after(int(sleep_time / 1000000), callback)
        else:
            print("sleeping until woken up")



if __name__ == "__main__":
    toplevel = tkinter.Tk()
    toplevel.resizable(0,0)

    # For some reason, we have to initialize all 3 statecharts in code (instead of just the 'system' statechart)
    sc = statechart.Statechart()

    controller = Controller()

    sc.timer_service = TimerService(controller)

    def update_callback():
        string_simtime.set('{:.3f}'.format((controller.simulated_time - controller.start_time) / 1000000000))
    sim = RealTimeSimulation(controller, toplevel, update_callback, scale=10.0)

    toplevel.title("Traffic Light")

    canvas = tkinter.Canvas(toplevel, bd=0, bg='#000000',
            width=100,
            height=300)

    OFF_COLOR = '#333333'

    red = canvas.create_oval(10,10,90,90, fill=OFF_COLOR)
    yellow = canvas.create_oval(10,110,90,190, fill=OFF_COLOR)
    green = canvas.create_oval(10,210,90,290, fill=OFF_COLOR)

    canvas.pack()

    tkinter.Button(toplevel, text="Police interrupt", command=lambda:sim.add_input(sc, "interrupt")).pack()
    tkinter.Button(toplevel, text="On / off", command=lambda:sim.add_input(sc, "toggle_on_off")).pack()

    def displayNone():
        canvas.itemconfigure(red, fill=OFF_COLOR)
        canvas.itemconfigure(yellow, fill=OFF_COLOR)
        canvas.itemconfigure(green, fill=OFF_COLOR)

    def displayRed():
        displayNone()
        canvas.itemconfigure(red, fill='#ff2222')

    def displayYellow():
        displayNone()
        canvas.itemconfigure(yellow, fill='#ffff22')

    def displayRedAndYellow():
        displayNone()
        canvas.itemconfigure(red, fill='#ff2222')
        canvas.itemconfigure(yellow, fill='#ffff22')
        

    def displayGreen():
        displayNone()
        canvas.itemconfigure(green, fill='#22ff22')


    class CallbackObserver(Observer):
        def __init__(self, callback):
            self.callback = callback

        def next(self, value=None):
            if value is None:
                self.callback()
            else:
                self.callback(value)

    sc.display_red_observable.subscribe(CallbackObserver(displayRed))
    sc.display_yellow_observable.subscribe(CallbackObserver(displayYellow))
    sc.display_green_observable.subscribe(CallbackObserver(displayGreen))
    sc.display_none_observable.subscribe(CallbackObserver(displayNone))

    string_simtime = tkinter.StringVar()
    string_counter = tkinter.StringVar()

    class CounterObserver(Observer):
        def next(self, value):
            if value == -1:
                string_counter.set('')
            else:
                string_counter.set(str(value))

    sc.set_timer_value_observable.subscribe(CounterObserver())

    sim_frame = tkinter.Frame(toplevel)
    tkinter.Label(sim_frame, text="simulated time").grid(column=0, row=0, columnspan=2)
    tkinter.Entry(sim_frame, state='readonly', width=8, textvariable=string_simtime, justify=tkinter.RIGHT).grid(row=1, column=0)
    tkinter.Label(sim_frame, text="s").grid(row=1, column=1)
    sim_frame.pack(side=tkinter.RIGHT)


    count_frame = tkinter.Frame(toplevel)
    tkinter.Label(count_frame, text="remaining").grid(column=0, row=0, columnspan=2)
    tkinter.Entry(count_frame, state='readonly', width=8, textvariable=string_counter, justify=tkinter.RIGHT).grid(row=1, column=0)
    tkinter.Label(count_frame, text="s").grid(row=1, column=1)
    count_frame.pack(side=tkinter.RIGHT)


    # Enter default states
    controller.start(scaled_now())

    sc.enter()

    sim.interrupt() # schedule first wakeup
    toplevel.mainloop()
